#!/usr/bin/env python

# import logging

# from zimscraperlib.logging import getLogger

# logger = getLogger(__name__, level=logging.DEBUG)

# TMP_FOLDER = "tmp"

# VERSION = "0.0.1"

# #LANGS = ["fr", "en", "de", "it", "es", "pt", "tr", "jp", "nl", "ru", "zh", "ko"]
# LANGS = ["en"]



# -*- coding: utf-8 -*-

import pathlib

# import re
import tempfile
import urllib.parse
from dataclasses import dataclass, field
from typing import List, Optional, Set

from zimscraperlib.i18n import get_language_details

ROOT_DIR = pathlib.Path(__file__).parent
NAME = ROOT_DIR.name
DEFAULT_HOMEPAGE = "Main-Page"
# MAX_HTTP_404_THRESHOLD = 200

with open(ROOT_DIR.joinpath("VERSION"), "r") as fh:
    VERSION = fh.read().strip()

SCRAPER = f"{NAME} {VERSION}"

IMAGES_ENCODER_VERSION = 1
# VIDEOS_ENCODER_VERSION = 1
URLS = {
    "en": "https://www.ifixit.com",
    "pt": "https://pt.ifixit.com",
}

DEFAULT_GUIDE_IMAGE_URL = (
    "https://assets.cdn.ifixit.com/static/images/"
    "default_images/GuideNoImage_300x225.jpg"
)

DIFFICULTY_VERY_EASY = [
    "Very easy",
    "Muito fácil",
    "Très facile",
    "Sehr einfach",
    "Muy fácil",
    "Molto facile",
    "Çok kolay",
    "とても簡単",
    "Zeer eenvoudig",
    "Очень просто",
    "아주 쉬움",
    "非常容易",
]  # guide 219
DIFFICULTY_EASY = [
    "Easy",
    "Fácil",
    "Facile",
    "Einfach",
    "Fácil",
    "Facile",
    "Kolay",
    "簡単",
    "Eenvoudig",
    "Просто",
    "쉬움",
    "简单",
]  # guide 220
DIFFICULTY_MODERATE = [
    "Moderate",
    "Moderado",
    "Modérée",
    "Mittel",
    "Moderado",
    "Moderato",
    "Orta",
    "普通",
    "Gemiddeld",
    "Средняя",
    "보통",
    "中等",
]  # guide 202
DIFFICULTY_HARD = [
    "Difficult",
    "Difícil",
    "Difficile",
    "Schwierig",
    "Difícil",
    "Difficile",
    "Zor",
    "難しい",
    "Moeilijk",
    "Сложно",
    "어려움",
    "困难",
]  # guide 206
DIFFICULTY_VERY_HARD = [
    "Very difficult",
    "Muito difícil",
    "Très difficile",
    "Sehr schwierig",
    "Muy difícil",
    "Molto difficile",
    "Çok zor",
    "とても難しい",
    "Zeer moeilijk",
    "Очень сложно",
    "매우 어려움",
    "非常困难",
]  # guide 46465
CATEGORY_LABELS = {
    "en": {
        "author": "Author: ",
        "categories": " Categories",
        "featured_guides": "Featured Guides",
        "technique_guides": "Techniques",
        "replacement_guides": "Replacement Guides",
        "teardown_guides": "Teardowns",
        "related_pages": "Related Pages",
        "in_progress_guides": "In Progress Guides",
        "disassembly_guides": "Disassembly Guides",
        "repairability": "Repairability:",
    },
    "pt": {
        "author": "Autor: ",
        "categories": " categorias",
        "featured_guides": "Guia em destaque",
        "technique_guides": "Técnicas",
        "replacement_guides": "Guias de reposição",
        "teardown_guides": "Teardowns",
        "related_pages": "Páginas relacionadas",
        "in_progress_guides": "Guias em andamento",
        "disassembly_guides": "Guias de Desmontagem",
        "repairability": "Reparabilidade:",
    },
}
# https://pt.ifixit.com/Device/Mac
# https://pt.ifixit.com/Device/Apple_Watch
# https://pt.ifixit.com/Device/Logitech__G502_Hero
# https://pt.ifixit.com/Guide/MacBook+Air+11-Inch+Late+2010+Battery+Replacement/4384

GUIDE_LABELS = {
    "en": {
        "written_by": "Written By:",
        "difficulty": "Difficulty",
        "steps": "Steps",
        "time_required": " Time Required",
        "sections": "Sections",
        "flags": "Flags",
        "introduction": "Introduction",
        "step_no": "Step ",
        "conclusion": "Conclusion",
        "author": "Author",
        "reputation": "Reputation",
        "member_since": "Member since: ",
        "published": "Published: ",
        "teardown": "Teardown",
    },
    "fr": {
        "written_by": "Rédigé par :",
        "difficulty": "Difficulté",
        "steps": "Étapes",
        "time_required": "Temps nécessaire",
        "sections": "Sections",
        "flags": "Drapeaux",
        "introduction": "Introduction",
        "step_no": "Étape ",
        "conclusion": "Conclusion",
        "author": "Auteur",
        "reputation": "Réputation",
        "member_since": "Membre depuis le ",
        "published": "Publication : ",
        "teardown": "Vue éclatée",
    },
    "pt": {
        "written_by": "Escrito Por:",
        "difficulty": "Dificuldade",
        "steps": "Passos",
        "time_required": "Tempo necessário",
        "sections": "Partes",
        "flags": "Sinalizadores",
        "introduction": "Introdução",
        "step_no": "Passo ",
        "conclusion": "Conclusão",
        "author": "Autor(a)",
        "reputation": "Reputação",
        "member_since": "Membro desde: ",
        "published": "Publicado em: ",
        "teardown": "Teardown",
    },
}

# https://pt.ifixit.com/Teardown/Apple+Watch+Teardown/40655

API_PREFIX = "/api/2.0"


@dataclass
class Conf:
    required = [
        "lang_code",
        "output_dir",
    ]

    lang_code: str = ""
    language: dict = field(default_factory=dict)
    main_url: str = ""

    # zim params
    name: str = ""
    title: Optional[str] = ""
    description: Optional[str] = ""
    author: Optional[str] = ""
    publisher: Optional[str] = ""
    fname: Optional[str] = ""
    tag: List[str] = field(default_factory=list)

    # customization
    icon: Optional[str] = ""
    categories: Set[str] = field(default_factory=set)

    # filesystem
    _output_dir: Optional[str] = "."
    _tmp_dir: Optional[str] = "."
    output_dir: Optional[pathlib.Path] = None
    tmp_dir: Optional[pathlib.Path] = None

    # performances
    nb_threads: Optional[int] = -1
    s3_url_with_credentials: Optional[str] = ""

    #     # quality
    #     without_videos: Optional[bool] = False
    #     without_external_links: Optional[bool] = False
    #     exclude: Optional[str] = ""
    #     only: Optional[str] = ""
    #     low_quality: Optional[bool] = False
    #     video_format: Optional[str] = "webm"

    # debug/devel
    build_dir_is_tmp_dir: Optional[bool] = False
    keep_build_dir: Optional[bool] = False
    debug: Optional[bool] = False
    delay: Optional[float] = 0
    api_delay: Optional[float] = 0
    cdn_delay: Optional[float] = 0
    stats_filename: Optional[str] = None
    skip_checks: Optional[bool] = False
    #     skip_footer_links: Optional[bool] = False
    #     single_article: Optional[str] = ""
    #     full_mode: Optional[bool] = False
    #     single_category: Optional[str] = None

    @staticmethod
    def get_url(lang_code: str) -> urllib.parse.ParseResult:
        return urllib.parse.urlparse(URLS[lang_code])

    @property
    def domain(self) -> str:
        return self.main_url.netloc

    @property
    def api_url(self) -> str:
        return self.main_url + API_PREFIX

    @property
    def s3_url(self) -> str:
        return self.s3_url_with_credentials

    def __post_init__(self):
        self.main_url = Conf.get_url(self.lang_code)
        self.language = get_language_details(self.lang_code)
        self.output_dir = pathlib.Path(self._output_dir).expanduser().resolve()
        self.output_dir.mkdir(parents=True, exist_ok=True)

        self.tmp_dir = pathlib.Path(self._tmp_dir).expanduser().resolve()
        self.tmp_dir.mkdir(parents=True, exist_ok=True)
        if self.build_dir_is_tmp_dir:
            self.build_dir = self.tmp_dir
        else:
            self.build_dir = pathlib.Path(
                tempfile.mkdtemp(prefix=f"ifixit_{self.lang_code}_", dir=self.tmp_dir)
            )
        # self.build_dir.joinpath("videos").mkdir(parents=True, exist_ok=True)

        if self.stats_filename:
            self.stats_filename = pathlib.Path(self.stats_filename).expanduser()
            self.stats_filename.parent.mkdir(parents=True, exist_ok=True)

        # support semi-colon separated tags as well
        if self.tag:
            for tag in self.tag.copy():
                if ";" in tag:
                    self.tag += [p.strip() for p in tag.split(";")]
                    self.tag.remove(tag)

        self.categories = set() if self.categories is None else self.categories


#         # the solely requested category or None
#         self.single_category = (
#             re.sub(r"/$", "", list(self.categories)[0])
#             if len(self.categories) == 1
#             else None
#         )
#         # whether requesting a _full mode_ (complete wiki)
#         self.full_mode = not self.categories and not self.only and not self.exclude



# -*- coding: utf-8 -*-

import json
import pathlib
import re
import shutil
from datetime import datetime

from jinja2 import Environment, FileSystemLoader, select_autoescape

from .constants import (
    CATEGORY_LABELS,
    DEFAULT_HOMEPAGE,
    DIFFICULTY_EASY,
    DIFFICULTY_HARD,
    DIFFICULTY_MODERATE,
    DIFFICULTY_VERY_EASY,
    DIFFICULTY_VERY_HARD,
    GUIDE_LABELS,
    ROOT_DIR,
    Conf,
)
from .shared import Global, GlobalMixin, logger
from .utils import (
    get_api_content,
    get_image_path,
    get_image_url,
    get_soup,
    guides_in_progress,
    setlocale,
    setup_s3_and_check_credentials,
)


class CategoryHomePageContentError(Exception):
    pass


class UnexpectedDataKindException(Exception):
    pass


class ifixit2zim(GlobalMixin):
    def __init__(self, **kwargs):

        Global.conf = Conf(**kwargs)
        for option in Global.conf.required:
            if getattr(Global.conf, option) is None:
                raise ValueError(f"Missing parameter `{option}`")

        # jinja2 environment setup
        self.env = Environment(
            loader=FileSystemLoader(ROOT_DIR.joinpath("templates")),
            autoescape=select_autoescape(),
        )
        # self.env.filters["digest"] = get_digest
        self.env.filters["guides_in_progress"] = guides_in_progress
        self.env.filters["get_image_path"] = get_image_path
        self.env.filters["get_image_url"] = get_image_url

        # jinja context that we'll pass to all templates
        self.env_context = {"conf": Global.conf}

        self.category_template = self.env.get_template("category.html")
        self.guide_template = self.env.get_template("guide.html")

        # used to prevent twice processing the resources of CSS
        # that we are going to download as they link to each other
        # Source HTML references a dynamic CSS that is built using a varietyof
        # features so it's very common different CSS urls references the same
        # resources (imgs)
        self.resources_digests = set()
        # List of URLs which returned HTTP 404.
        # There are legit scenarios for 404 on wikiHow: login pages
        # we need to track them for later use
        self.missing_articles = set()
        self.missing_categories = set()

    @property
    def build_dir(self):
        return self.conf.build_dir

    def cleanup(self):
        """Remove temp files and release resources before exiting"""
        if not self.conf.keep_build_dir:
            logger.debug(f"Removing {self.build_dir}")
            shutil.rmtree(self.build_dir, ignore_errors=True)

    def get_online_metadata(self):
        """metadata from online website, looking at homepage source code"""
        logger.info("Fetching website metadata")

        soup, _ = get_soup("/")

        return {
            "title": soup.find("title").string,
            "description": soup.find("meta", attrs={"name": "description"}).attrs.get(
                "content"
            ),
            "footer_stats": self._extract_footer_stats_from_page(soup),
            "footer_copyright": self._extract_footer_copyright_from_page(soup),
        }

    def sanitize_inputs(self):
        """input & metadata sanitation"""
        logger.debug("Checking user-provided metadata")

        if not self.conf.name:
            self.conf.name = "ifixit_{lang}_{selection}".format(
                lang=self.conf.language["iso-639-1"],
                selection="selection" if self.conf.categories else "all",
            )

        period = datetime.now().strftime("%Y-%m")
        if self.conf.fname:
            # make sure we were given a filename and not a path
            self.conf.fname = pathlib.Path(self.conf.fname.format(period=period))
            if pathlib.Path(self.conf.fname.name) != self.conf.fname:
                raise ValueError(f"filename is not a filename: {self.conf.fname}")
        else:
            self.conf.fname = f"{self.conf.name}_{period}.zim"

        if not self.conf.title:
            self.conf.title = self.metadata["title"]
        self.conf.title = self.conf.title.strip()

        if not self.conf.description:
            self.conf.description = self.metadata["description"]
        self.conf.description = self.conf.description.strip()

        if not self.conf.author:
            self.conf.author = "iFixit"
        self.conf.author = self.conf.author.strip()

        if not self.conf.publisher:
            self.conf.publisher = "openZIM"
        self.conf.publisher = self.conf.publisher.strip()

        self.conf.tags = list(
            set(
                self.conf.tag
                + ["_category:ifixit", "ifixit", "_videos:yes", "_pictures:yes"]
            )
        )

        logger.debug(
            "Configuration after sanitization:\n"
            f"name: {self.conf.name}\n"
            f"fname: {self.conf.fname}\n"
            f"name: {self.conf.author}\n"
            f"fname: {self.conf.publisher}"
        )

    def add_assets(self):
        """download and add site-wide assets, identified in metadata step"""
        logger.info("Adding assets")

        # recursively add our assets, at a path identical to position in repo
        assets_root = pathlib.Path(ROOT_DIR.joinpath("assets"))
        for fpath in assets_root.glob("**/*"):
            if not fpath.is_file():
                continue
            path = str(fpath.relative_to(ROOT_DIR))

            logger.debug(f"> {path}")
            with self.lock:
                self.creator.add_item_for(path=path, fpath=fpath)

    def _process_categories(self, categories):
        for category in categories:
            self.expected_categories.add(category)
            self._process_categories(categories[category])

    def build_expected_categories(self):
        logger.info("Downloading categories")
        categories = get_api_content("/categories", includeStubs=True)
        self._process_categories(categories)
        logger.info("{} categories found".format(len(self.expected_categories)))

    guide_regex_full = re.compile(
        r"href=\"https://\w*\.ifixit\.\w*/Guide/.*/(?P<guide_id>\d*)\""
    )
    guide_regex_rel = re.compile(r"href=\"/Guide/.*/(?P<guide_id>\d*).*?\"")
    content_image_regex = re.compile(
        r"(?P<prefix>https://guide-images\.cdn\.ifixit\.com/igi/)"
        r"(?P<image_filename>\w*)\.\w*"
    )
    image_regex = re.compile(r"<img(?P<before>.*?)src=\"(?P<url>.*?)\"")
    device_link_regex_without_href = re.compile(r"/Device/(?P<device>.*)")
    device_link_regex_with_href = re.compile(r"href=\"/Device/(?P<device>.*)\"")

    def _get_image_guid_from_src(self, src):
        # return src
        return self.content_image_regex.sub("\\g<image_filename>", src)

    def _get_category_from_href(self, href):
        return self.device_link_regex_without_href.sub('\\g<device>"', href)

    def _extract_page_title_from_page(self, soup):
        page_title_selector = "h1.page-title span"
        p = soup.select(page_title_selector)
        if len(p) == 0:
            raise CategoryHomePageContentError(
                "No text found in page with selector " f"'{page_title_selector}'"
            )
        if len(p) > 1:
            raise CategoryHomePageContentError(
                "Too many text found in page with selector " f"'{page_title_selector}'"
            )
        text = p[0].text
        if len(text) == 0:
            raise CategoryHomePageContentError(
                "Empty text found in page with selector " f"'{page_title_selector}'"
            )
        return text

    def _extract_primary_title_from_page(self, soup):
        primary_title_selector = "div.primary-divider p"
        p = soup.select(primary_title_selector)
        if len(p) == 0:
            raise CategoryHomePageContentError(
                "No text found in page with selector " f"'{primary_title_selector}'"
            )
        if len(p) > 1:
            raise CategoryHomePageContentError(
                "Too many text found in page with selector "
                f"'{primary_title_selector}'"
            )
        text = p[0].text
        if len(text) == 0:
            raise CategoryHomePageContentError(
                "Empty text found in page with selector " f"'{primary_title_selector}'"
            )
        return text

    def _extract_secondary_title_from_page(self, soup):
        secondary_title_selector = "div.secondary-divider p"
        p = soup.select(secondary_title_selector)
        if len(p) == 0:
            raise CategoryHomePageContentError(
                "No text found in page with selector " f"'{secondary_title_selector}'"
            )
        if len(p) > 1:
            raise CategoryHomePageContentError(
                "Too many text found in page with selector "
                f"'{secondary_title_selector}'"
            )
        text = p[0].text
        if len(text) == 0:
            raise CategoryHomePageContentError(
                "Empty text found in page with selector "
                f"'{secondary_title_selector}'"
            )
        return text

    def _extract_callout_from_page(self, soup):
        return {
            "content": self._extract_callout_content_from_page(soup),
            "img_url": self._extract_callout_img_src_from_page(soup),
        }

    def _extract_callout_content_from_page(self, soup):
        page_callout_selector = "div.page-callout-content"
        p = soup.select(page_callout_selector)
        if len(p) == 0:
            raise CategoryHomePageContentError(
                "No callout content found in page with selector "
                f"'{page_callout_selector}'"
            )
        if len(p) > 1:
            raise CategoryHomePageContentError(
                "Too many callout content found in page with selector "
                f"'{page_callout_selector}'"
            )
        return f"{p[0]}"

    def _extract_callout_img_src_from_page(self, soup):
        page_callout_selector = "div.page-callout-inner img"
        p = soup.select(page_callout_selector)
        if len(p) == 0:
            raise CategoryHomePageContentError(
                "No callout img found in page with selector "
                f"'{page_callout_selector}'"
            )
        if len(p) > 1:
            raise CategoryHomePageContentError(
                "Too many callout img found in page with selector "
                f"'{page_callout_selector}'"
            )
        src = p[0].attrs.get("src")
        if len(src) == 0:
            raise CategoryHomePageContentError(
                "Empty img src found in featured callout with selector "
                f"'{page_callout_selector}'"
            )
        return src

    def _extract_featured_categories_from_page(self, soup):
        featured_categories_css_selector = "a.featured-category-item"
        featured_categories = [
            {
                "text": self._extract_text_from_featured_category(fc),
                "img_url": self._extract_img_src_from_featured_category(fc),
                "name": self._extract_name_from_featured_category(fc),
                "title": self._extract_title_from_featured_category(fc),
            }
            for fc in soup.select(featured_categories_css_selector)
        ]

        if len(featured_categories) == 0:
            raise CategoryHomePageContentError(
                "No featured categories found with selector "
                f"'{featured_categories_css_selector}'"
            )
        return featured_categories

    def _extract_text_from_featured_category(self, fc):
        featured_category_text_css_selector = "p.featured-category-title"
        p = fc.select(featured_category_text_css_selector)
        if len(p) == 0:
            raise CategoryHomePageContentError(
                "No text found in featured category with selector "
                f"'{featured_category_text_css_selector}'"
            )
        if len(p) > 1:
            raise CategoryHomePageContentError(
                "Too many text found in featured category with selector "
                f"'{featured_category_text_css_selector}'"
            )
        text = p[0].text
        if len(text) == 0:
            raise CategoryHomePageContentError(
                "Empty text found in featured category with selector "
                f"'{featured_category_text_css_selector}'"
            )
        return text

    def _extract_img_src_from_featured_category(self, fc):
        featured_category_img_css_selector = "img"
        p = fc.select(featured_category_img_css_selector)
        if len(p) == 0:
            raise CategoryHomePageContentError(
                "No img found in featured category with selector "
                f"'{featured_category_img_css_selector}'"
            )
        if len(p) > 1:
            raise CategoryHomePageContentError(
                "Too many img found in featured category with selector "
                f"'{featured_category_img_css_selector}'"
            )
        src = p[0].attrs.get("src")
        if len(src) == 0:
            raise CategoryHomePageContentError(
                "Empty img src found in featured category with selector "
                f"'{featured_category_img_css_selector}'"
            )
        return src

    def _extract_name_from_featured_category(self, fc):
        href = fc.attrs.get("href")
        if len(href) == 0:
            raise CategoryHomePageContentError("Empty href found in featured category")
        name = self.device_link_regex_without_href.sub("\\g<device>", href)
        if name == href:
            raise CategoryHomePageContentError(
                f"Extracting name from featured category failed ; href:'{href}'"
            )
        return name

    def _extract_title_from_featured_category(self, fc):
        title = fc.attrs.get("title")
        if len(title) == 0:
            raise CategoryHomePageContentError("Empty title found in featured category")
        return title

    def _extract_sub_categories_from_page(self, soup):
        sub_categories_css_selector = "a.sub-category"
        sub_categories = [
            {
                "text": self._extract_text_from_sub_category(fc),
                "name": self._extract_name_from_sub_category(fc),
                "count": self._extract_count_from_sub_category(fc),
                "title": self._extract_title_from_sub_category(fc),
            }
            for fc in soup.select(sub_categories_css_selector)
        ]
        if len(sub_categories) == 0:
            raise CategoryHomePageContentError(
                "No sub-categories found with selector "
                f"'{sub_categories_css_selector}'"
            )
        return sub_categories

    def _extract_text_from_sub_category(self, sc):
        sub_category_text_css_selector = "span.sub-category-title-text"
        p = sc.select(sub_category_text_css_selector)
        if len(p) == 0:
            raise CategoryHomePageContentError(
                "No text found in sub-category with selector "
                f"'{sub_category_text_css_selector}'"
            )
        if len(p) > 1:
            raise CategoryHomePageContentError(
                "Too many text found in sub-category with selector "
                f"'{sub_category_text_css_selector}'"
            )
        text = p[0].text
        if len(text) == 0:
            raise CategoryHomePageContentError(
                "Empty text found in sub-category with selector "
                f"'{sub_category_text_css_selector}'"
            )
        return text

    def _extract_name_from_sub_category(self, sc):
        href = sc.attrs.get("href")
        if len(href) == 0:
            raise CategoryHomePageContentError("Empty href found in sub-category")
        name = self.device_link_regex_without_href.sub("\\g<device>", href)
        if name == href:
            raise CategoryHomePageContentError(
                f"Extracting name from sub-category failed ; href:'{href}'"
            )
        return name

    def _extract_count_from_sub_category(self, sc):
        sub_category_img_css_selector = "span.overflow-slide-in"
        p = sc.select(sub_category_img_css_selector)
        if len(p) == 0:
            raise CategoryHomePageContentError(
                "No span found in sub-category with selector "
                f"'{sub_category_img_css_selector}'"
            )
        if len(p) > 1:
            raise CategoryHomePageContentError(
                "Too many span found in sub-category with selector "
                f"'{sub_category_img_css_selector}'"
            )
        text = p[0].text
        if len(text) == 0:
            raise CategoryHomePageContentError(
                "Empty span text found in sub-category with selector "
                f"'{sub_category_img_css_selector}'"
            )
        try:
            return int(text)
        except ValueError:
            raise CategoryHomePageContentError(
                f"Failed to convert span text '{text}' to integer for " "sub-category"
            )

    def _extract_title_from_sub_category(self, sc):
        sub_category_img_css_selector = "span.overflow-slide-in"
        p = sc.select(sub_category_img_css_selector)
        if len(p) == 0:
            raise CategoryHomePageContentError(
                "No span found in sub-category with selector "
                f"'{sub_category_img_css_selector}'"
            )
        if len(p) > 1:
            raise CategoryHomePageContentError(
                "Too many span found in sub-category with selector "
                f"'{sub_category_img_css_selector}'"
            )
        title = p[0].attrs.get("title")
        if len(title) == 0:
            raise CategoryHomePageContentError(
                "Empty span title found in sub-category with selector "
                f"'{sub_category_img_css_selector}'"
            )
        return title

    def _extract_footer_stats_from_page(self, soup):
        footer_stats_css_selector = "div.footer-stats div"
        footer_stats = [
            self._extract_details_from_footer_stats(fc)
            for fc in soup.select(footer_stats_css_selector)
        ]
        if len(footer_stats) == 0:
            raise CategoryHomePageContentError(
                "No footer stats found with selector " f"'{footer_stats_css_selector}'"
            )
        return footer_stats

    def _extract_details_from_footer_stats(self, fs):
        footer_stats_text_css_selector = "p"
        p = fs.select(footer_stats_text_css_selector)
        if len(p) == 0:
            raise CategoryHomePageContentError(
                "No text found in footer stat with selector "
                f"'{footer_stats_text_css_selector}'"
            )
        if len(p) == 1:
            raise CategoryHomePageContentError(
                "Insufficient text found in footer stat with selector "
                f"'{footer_stats_text_css_selector}'"
            )
        if len(p) > 2:
            raise CategoryHomePageContentError(
                "Too many text found in footer stat with selector "
                f"'{footer_stats_text_css_selector}'"
            )
        text0 = p[0].text
        if len(text0) == 0:
            raise CategoryHomePageContentError(
                "Empty text found first paragraph of footer stat "
                f"'{footer_stats_text_css_selector}'"
            )
        text0digits = re.sub("[^0-9]", "", text0)
        if len(text0digits) == 0:
            raise CategoryHomePageContentError(
                f"No digits found in text '{text0}' of footer stat"
            )
        text1 = p[1].text
        if len(text1) == 0:
            raise CategoryHomePageContentError(
                "Empty text found second paragraph of footer stat "
                f"'{footer_stats_text_css_selector}'"
            )
        try:
            return {
                "value": int(text0digits),
                "text": text1,
            }
        except ValueError:
            raise CategoryHomePageContentError(
                f"Failed to convert text '{text0digits}' to integer for footer " "stat"
            )

    def _extract_footer_copyright_from_page(self, soup):
        footer_copyright_css_selector = "div.footer-container div.copyright"
        footer_copyright = soup.select(footer_copyright_css_selector)
        if len(footer_copyright) == 0:
            raise CategoryHomePageContentError(
                "No footer copyright found with selector "
                f"'{footer_copyright_css_selector}'"
            )
        main_footer_copyright = None
        for fc in footer_copyright:
            if len(fc.text.strip()) > 0:
                if main_footer_copyright is None:
                    main_footer_copyright = fc
                else:
                    raise CategoryHomePageContentError(
                        "Too many footer copyright with selector "
                        f"'{footer_copyright_css_selector}'"
                    )
        if main_footer_copyright is None:
            raise CategoryHomePageContentError(
                "Empty footer copyright found with selector "
                f"'{footer_copyright_css_selector}'"
            )

        return f"{main_footer_copyright}"

    def scrape_homepage(self):

        logger.info("Scraping homepage")

        soup, _ = get_soup("/Guide")

        logger.debug("Processing homepage")

        # extract and clean main content
        home_content = {
            "main_title": soup.find("title").string,
            "page_title": self._extract_page_title_from_page(soup),
            "primary_title": self._extract_primary_title_from_page(soup),
            "secondary_title": self._extract_secondary_title_from_page(soup),
            "callout": self._extract_callout_from_page(soup),
            "featured_categories": self._extract_featured_categories_from_page(soup),
            "sub_categories": self._extract_sub_categories_from_page(soup),
        }

        for fc in home_content["featured_categories"]:
            image_path = Global.imager.defer(url=fc["img_url"])
            fc["img_path"] = image_path

        home_content["callout"]["img_path"] = Global.imager.defer(
            url=home_content["callout"]["img_url"]
        )

        logger.debug(
            "Content extracted from /Guide:\n" f"{json.dumps(home_content,indent=2)}"
        )

        page = self.env.get_template("home.html").render(
            home_content=home_content, metadata=self.metadata
        )

        with self.lock:
            self.creator.add_item_for(
                path="home/home.html",
                title=self.conf.title,
                content=page,
                mimetype="text/html",
                is_front=True,
            )

            self.creator.add_redirect(
                path=DEFAULT_HOMEPAGE, target_path="home/home.html"
            )

    def scrape_one_category(self, category):
        category_content = get_api_content(
            f"/wikis/CATEGORY/{category}", langid=self.conf.lang_code
        )

        logger.debug(f"Processing category {category}")

        for guide in category_content["featured_guides"]:
            if guide["type"] not in [
                "replacement",
                "technique",
                "teardown",
                "disassembly",
            ]:
                raise UnexpectedDataKindException(
                    "Unsupported type of guide: {} for featured_guide {}".format(
                        guide["type"], guide["guideid"]
                    )
                )
            else:
                self.expected_guides[guide["guideid"]] = {
                    "guideid": guide["guideid"],
                    "locale": guide["locale"],
                }
        for guide in category_content["guides"]:
            if guide["type"] not in [
                "replacement",
                "technique",
                "teardown",
                "disassembly",
            ]:
                raise UnexpectedDataKindException(
                    "Unsupported type of guide: {} for guide {}".format(
                        guide["type"], guide["guideid"]
                    )
                )
            else:
                self.expected_guides[guide["guideid"]] = {
                    "guideid": guide["guideid"],
                    "locale": guide["locale"],
                }

        def _replace_image(m):
            orig_url = m.group("url")
            new_url = get_image_path(orig_url)
            return f"<img{m.group('before')}src=\"{new_url}\""

        category_content["contents_rendered"] = re.sub(
            self.image_regex, _replace_image, category_content["contents_rendered"]
        )
        category_content["contents_rendered"] = self.device_link_regex_with_href.sub(
            'href="./category_\\g<device>.html"',
            category_content["contents_rendered"],
        )
        category_content["filename"] = re.sub(r"\s", "_", category_content["title"])
        for idx, child in enumerate(category_content["children"]):
            category_content["children"][idx]["filename"] = re.sub(
                r"\s",
                "_",
                category_content["children"][idx]["title"],
            )
        category_rendered = self.category_template.render(
            category=category_content,
            label=CATEGORY_LABELS[self.conf.lang_code],
            metadata=self.metadata,
            lang=self.conf.lang_code,
        )
        with self.lock:
            self.creator.add_item_for(
                path=f"categories/category_{category_content['filename']}.html",
                title=category_content["display_title"],
                content=category_rendered,
                mimetype="text/html",
                is_front=True,
            )

    def scrape_categories(self):

        logger.info(f"Scraping {len(self.expected_categories)} categories")

        num_category = 1
        for category in self.expected_categories:
            # if category not in ["Mac", "Tablet", "Apple Pro Speakers M8756"]:
            #     continue
            logger.info(
                f"Scraping category {category} ({num_category}/"
                "{len(self.expected_categories)})"
            )
            self.scrape_one_category(category)
            num_category += 1

    def scrape_one_guide(self, guide):
        guide_content = get_api_content(
            f"/guides/{guide['guideid']}", langid=guide["locale"]
        )

        logger.debug(f"Processing guide {guide['guideid']}")

        if guide_content["type"] != "teardown":
            if guide_content["difficulty"] in DIFFICULTY_VERY_EASY:
                guide_content["difficulty_class"] = "difficulty-1"
            elif guide_content["difficulty"] in DIFFICULTY_EASY:
                guide_content["difficulty_class"] = "difficulty-2"
            elif guide_content["difficulty"] in DIFFICULTY_MODERATE:
                guide_content["difficulty_class"] = "difficulty-3"
            elif guide_content["difficulty"] in DIFFICULTY_HARD:
                guide_content["difficulty_class"] = "difficulty-4"
            elif guide_content["difficulty"] in DIFFICULTY_VERY_HARD:
                guide_content["difficulty_class"] = "difficulty-5"
            else:
                raise UnexpectedDataKindException(
                    "Unknown guide difficulty: '{}' in guide {}".format(
                        guide_content["difficulty"],
                        guide_content["guideid"],
                    )
                )
        with setlocale("en_GB"):
            if guide_content["author"]["join_date"]:
                guide_content["author"]["join_date_rendered"] = datetime.strftime(
                    datetime.fromtimestamp(guide_content["author"]["join_date"]),
                    "%x",
                )
            # TODO: format published date as June 10, 2014 instead of 11/06/2014
            if guide_content["published_date"]:
                guide_content["published_date_rendered"] = datetime.strftime(
                    datetime.fromtimestamp(guide_content["published_date"]),
                    "%x",
                )
        guide_content["introduction_rendered"] = self.guide_regex_full.sub(
            'href="./guide_\\g<guide_id>.html"',
            guide_content["introduction_rendered"],
        )
        guide_content["introduction_rendered"] = self.guide_regex_rel.sub(
            'href="./guide_\\g<guide_id>.html"',
            guide_content["introduction_rendered"],
        )
        guide_content["conclusion_rendered"] = self.guide_regex_full.sub(
            'href="./guide_\\g<guide_id>.html"',
            guide_content["conclusion_rendered"],
        )
        guide_content["conclusion_rendered"] = self.guide_regex_rel.sub(
            'href="./guide_\\g<guide_id>.html"',
            guide_content["conclusion_rendered"],
        )
        for step in guide_content["steps"]:
            if not step["media"]:
                raise UnexpectedDataKindException(
                    "Missing media attribute in step {} of guide {}".format(
                        step["stepid"], guide_content["guideid"]
                    )
                )
            if step["media"]["type"] not in [
                "image",
                "video",
                "embed",
            ]:
                raise UnexpectedDataKindException(
                    "Unrecognized media type in step {} of guide {}".format(
                        step["stepid"], guide_content["guideid"]
                    )
                )
            if step["media"]["type"] == "video":
                if "data" not in step["media"] or not step["media"]["data"]:
                    raise UnexpectedDataKindException(
                        "Missing 'data' in step {} of guide {}".format(
                            step["stepid"], guide_content["guideid"]
                        )
                    )
                if (
                    "image" not in step["media"]["data"]
                    or not step["media"]["data"]["image"]
                ):
                    raise UnexpectedDataKindException(
                        "Missing outer 'image' in step {} of guide {}".format(
                            step["stepid"], guide_content["guideid"]
                        )
                    )
                if (
                    "image" not in step["media"]["data"]["image"]
                    or not step["media"]["data"]["image"]["image"]
                ):
                    raise UnexpectedDataKindException(
                        "Missing inner 'image' in step {} of guide {}".format(
                            step["stepid"], guide_content["guideid"]
                        )
                    )
            if step["media"]["type"] == "embed":
                if "data" not in step["media"] or not step["media"]["data"]:
                    raise UnexpectedDataKindException(
                        "Missing 'data' in step {} of guide {}".format(
                            step["stepid"], guide_content["guideid"]
                        )
                    )
                if (
                    "html" not in step["media"]["data"]
                    or not step["media"]["data"]["html"]
                ):
                    raise UnexpectedDataKindException(
                        "Missing 'html' in step {} of guide {}".format(
                            step["stepid"], guide_content["guideid"]
                        )
                    )
            for line in step["lines"]:
                if not line["bullet"] in [
                    "black",
                    "red",
                    "orange",
                    "yellow",
                    "green",
                    "blue",
                    "light_blue",
                    "violet",
                    "icon_note",
                    "icon_caution",
                    "icon_caution",
                    "icon_reminder",
                ]:
                    raise UnexpectedDataKindException(
                        "Unrecognized bullet '{}' in step {} of guide {}".format(
                            line["bullet"],
                            step["stepid"],
                            guide_content["guideid"],
                        )
                    )
                line["text_rendered"] = self.guide_regex_full.sub(
                    'href="./guide_\\g<guide_id>.html"',
                    line["text_rendered"],
                )
                line["text_rendered"] = self.guide_regex_rel.sub(
                    'href="./guide_\\g<guide_id>.html"',
                    line["text_rendered"],
                )
        guide_rendered = self.guide_template.render(
            guide=guide_content,
            label=GUIDE_LABELS[self.conf.lang_code],
            metadata=self.metadata,
        )
        with self.lock:
            self.creator.add_item_for(
                path=f"guides/guide_{guide_content['guideid']}.html",
                title=guide_content["title"],
                content=guide_rendered,
                mimetype="text/html",
                is_front=True,
            )

    def scrape_guides(self):

        logger.info(f"Scraping {len(self.expected_guides)} guides")

        num_guide = 1
        for guideid, guide in self.expected_guides.items():
            # if guideid not in [132168, 120579]:
            #     continue
            logger.info(
                f"Scraping guide {guideid} ({num_guide}/{len(self.expected_guides)})"
            )
            self.scrape_one_guide(guide)
            num_guide += 1

    def run(self):
        s3_storage = (
            setup_s3_and_check_credentials(self.conf.s3_url_with_credentials)
            if self.conf.s3_url_with_credentials
            else None
        )
        s3_msg = (
            f"\n"
            f"  using cache: {s3_storage.url.netloc} "
            f"with bucket: {s3_storage.bucket_name}"
            if s3_storage
            else ""
        )
        del s3_storage

        logger.info(
            f"Starting scraper with:\n"
            f"  language: {self.conf.language['english']}"
            f" ({self.conf.domain})\n"
            f"  output_dir: {self.conf.output_dir}\n"
            f"  build_dir: {self.build_dir}\n"
            f"{s3_msg}"
        )

        Global.metadata = self.get_online_metadata()
        logger.debug(
            f"Additional metadata scrapped online:\n"
            f"title: {self.metadata['title']}\n"
            f"description: {self.metadata['description']}\n"
            f"footer_stats: {self.metadata['footer_stats']}\n"
            f"footer_copyright: {self.metadata['footer_copyright']}\n"
        )
        self.sanitize_inputs()

        logger.debug("Starting Zim creation")
        Global.setup()
        self.creator.start()

        try:
            self.build_expected_categories()

            self.add_assets()
            self.scrape_homepage()
            self.scrape_categories()
            self.scrape_guides()

            logger.info("Awaiting images")
            Global.img_executor.shutdown()

        except Exception as exc:
            # request Creator not to create a ZIM file on finish
            self.creator.can_finish = False
            if isinstance(exc, KeyboardInterrupt):
                logger.error("KeyboardInterrupt, exiting.")
            else:
                logger.error(f"Interrupting process due to error: {exc}")
                logger.exception(exc)
            self.imager.abort()
            Global.img_executor.shutdown(wait=False)
            return 1
        else:
            logger.info("Finishing ZIM file")
            # we need to release libzim's resources.
            # currently does nothing but crash if can_finish=False
            # but that's awaiting impl. at libkiwix level
            with self.lock:
                self.creator.finish()
            logger.info(
                f"Finished Zim {self.creator.filename.name} "
                f"in {self.creator.filename.parent}"
            )
        finally:
            self.cleanup()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# vim: ai ts=4 sts=4 et sw=4 nu

# import collections
import io
import locale
import re
import threading
import urllib.parse
import zlib
from contextlib import contextmanager
from typing import Union  # , Iterable, List, Tuple

import backoff
import bs4

# import cssbeautifier
import requests
from kiwixstorage import KiwixStorage
from pif import get_public_ip

# from tld import get_fld
from zimscraperlib.download import _get_retry_adapter, stream_file

from .constants import API_PREFIX, DEFAULT_GUIDE_IMAGE_URL
from .shared import Global, logger

# nlink = collections.namedtuple("Link", ("path", "name", "title"))

LOCALE_LOCK = threading.Lock()


class ImageUrlNotFound(Exception):
    pass


def to_path(url: str) -> str:
    """Path-part of an URL, without leading slash"""
    return re.sub(r"^/", "", urllib.parse.urlparse(url).path)


def get_url(path: str, **params) -> str:
    """url-encoded in-source website url for a path"""
    params_str = f"?{urllib.parse.urlencode(params)}" if params else ""
    return f"{Global.conf.main_url.geturl()}{urllib.parse.quote(path)}{params_str}"


def get_url_raw(path: str):
    """in-source website url for a path, untainted"""
    return f"{Global.conf.main_url.geturl()}{path}"


def to_url(value: str) -> str:
    """resolved potentially relative url from in-source link"""
    return value if value.startswith("http") else get_url_raw(value)


def to_rel(url: str) -> Union[None, str]:
    """path from URL if on our main domain, else None"""
    uri = urllib.parse.urlparse(url)
    if uri.netloc != Global.conf.domain:
        return None
    return uri.path


def no_leading_slash(text: str) -> str:
    """text with leading slash removed if present"""
    return re.sub(r"^/", "", text)


def no_trailing_slash(text: str) -> str:
    """text with trailing slash removed if present"""
    return re.sub(r"/$", "", text)


def only_path_of(url: str):
    """normalized path part of an url"""
    return normalize_ident(urllib.parse.urlparse(url).path)


def fetch(path: str, **params) -> str:
    """(source text, actual_paths) of a path from source website

    actual_paths is amn ordered list of paths that were traversed to get to content.
    Without redirection, it should be a single path, equal to request
    Final, target path is always last"""
    session = requests.Session()
    session.mount("http", _get_retry_adapter(10))  # tied to http and https
    resp = session.get(get_url(path, **params), params=params)
    resp.raise_for_status()

    # we have params meaning we requested a page (?pg=xxx)
    # assumption: this must be a category page (so on same domain)
    # we thus need to use redirection target (which lost param) with params
    if params and resp.history:
        return fetch(only_path_of(resp.url), **params)
    return resp.text, [
        no_leading_slash(only_path_of(r.url)) for r in resp.history + [resp]
    ]


def get_soup_of(text: str, unwrap: bool = False):
    """an lxml soup of an HTML string"""
    soup = bs4.BeautifulSoup(text, "lxml")
    if unwrap:
        for elem in ("body", "html"):
            getattr(soup, elem).unwrap()
    return soup


# def get_footer_crumbs_from(soup: bs4.element.Tag) -> List[Tuple[str, str, str]]:
#     """List of (url, name and title) of footer breadcrumbs"""

#     crumbs = []
#     for link in soup.select("#footer_crumbs ul li a[href]"):
#         if not link.attrs.get("href"):
#             continue
#         # might or might not be a Category link
#         try:
#             cat_ident = cat_ident_for(link.attrs["href"])
#         except Exception:
#             cat_ident = None
#         if cat_ident is None or cat_ident in Global.expected_categories:
#             crumbs.append(
#                 nlink(link.attrs["href"][1:], link.string, link.attrs.get("title"))
#             )
#     return crumbs


# def get_footer_links_from(soup: bs4.element.Tag) -> List[Tuple[str, str, str]]:
#     """list of namedtuple(path, name, title) of footer links"""
#     links = []

#     fld = get_fld(Global.conf.main_url.geturl())

#     # Skip some links with no offline value
#     for link in soup.select("#footer_links ul li a"):
#         if link.attrs.get("href") in (
#             "#",
#             "https://blog.wikihow.com/",
#             "/wikiHow:Jobs",
#             "https://www.wikihow.com/wikiHow:Contribute",
#         ):
#             continue

#         path = None
#         if link.attrs.get("href"):
#             url = urllib.parse.urlparse(to_url(link.attrs["href"]))

#             # skip external URLs (if any)
#             if get_fld(url.geturl()) != fld:
#                 continue
#             path = normalize_ident(url.path)[1:]

#         links.append(nlink(path, link.string, link.attrs.get("title")))
#     return links


def get_soup(path: str, **params) -> bs4.BeautifulSoup:
    """an lxml soup of a path on source website"""
    content, paths = fetch(path, **params)
    return get_soup_of(content), paths


# def soup_link_finder(elem: bs4.element.Tag) -> bool:
#     """bs4's find_all-friendly selector for linked styles in wikiHow"""
#     return (
#         elem.name == "link"
#         and elem.attrs.get("href")
#         and (elem.attrs.get("as") == "style" or elem.attrs.get("rel") == "stylesheet")
#     )


def get_digest(url: str) -> str:
    """simple digest of an url for mapping purpose"""
    return str(zlib.adler32(url.encode("UTF-8")))


# def cat_ident_for(href: str) -> str:
#     """decoded name of a category from a link target"""
#     return normalize_ident(href).split(":", 1)[1]


# def fix_pagination_links(soup: bs4.element.Tag):
#     """Replace ?pg= to _pg= in pagination links"""
#     for a in soup.select("#large_pagination a[href]"):
#         a["href"] = a["href"].replace("?pg=", "_pg=")


# def get_categorylisting_url():
#     return normalize_ident(
#         urllib.parse.urlparse(requests.get(to_url("/Special:CategoryListing")).url).path
#     )[1:]


# def get_youtube_id_from(url: str) -> str:
#     """Youtube video Id from a youtube URL"""
#     uri = urllib.parse.urlparse(url)
#     if uri.path.startswith("/embed/"):
#         m = re.match(r"^/embed/(?P<id>[^/]+)", uri.path)
#         if m:
#             return m.groupdict().get("id")
#     if uri.path.startswith("/watch"):
#         return urllib.parse.parse_qs(uri.query).get("v", [None]).pop()


# def normalize_youtube_url(url: str) -> str:
#     """harmonize youtube-URL to use a single (public viewing) format

#     format: https://www.youtube.com/watch?v=C1vI8k-JEsQ"""

#     yid = get_youtube_id_from(url)
#     if yid:
#         return f"https://www.youtube.com/watch?v={yid}"
#     return url


def normalize_ident(ident: str) -> str:
    """URL-decoded category identifier"""
    return urllib.parse.unquote(ident)


# def article_ident_for(href: str) -> str:
#     """decoded name of an article from a link target"""
#     return normalize_ident(to_rel(href))[1:]


# def parse_css(style: str) -> Tuple[str, List[Tuple[str, str]]]:
#     """(css, resources) of transformed CSS string and resources list

#     reads a CSS string and returns it transformed
#     with url() replaced by offlinable path.

#     resources list is list of tuples, each containing the URL to get data from
#     and the path to store it at.
#     ex: ("http://goto.img/hello.png", "img/hello.png")"""

#     output = ""
#     resources = []

#     def write(line):
#         nonlocal output
#         output += line + "\n"

#     pattern = "url("
#     for line in cssbeautifier.beautify(style).split("\n"):
#         if pattern not in line:
#             write(line)
#             continue

#         start = line.index(pattern) + len(pattern)
#         end = line.index(")")

#         # check whether it's quoted or not
#         if line[start + 1] in ("'", '"'):
#             start += 1
#             end -= 1

#         url = line[start:end]

#         if url.startswith("data:"):
#             write(line)
#             continue

#         path = f"assets/{get_digest(url)}"
#         resources.append((url, path))
#         # resources are added on same level (assets/xxx) as css itself
#         write(line[0:start] + "../" + path + line[end:])

#     return output, resources


# def first(*args: Iterable[object]) -> object:
#     """first non-None value from *args ; fallback to empty string"""
#     return next((item for item in args if item is not None), "")


# def rebuild_uri(
#     uri: urllib.parse.ParseResult,
#     scheme: str = None,
#     username: str = None,
#     password: str = None,
#     hostname: str = None,
#     port: Union[str, int] = None,
#     path: str = None,
#     params: str = None,
#     query: str = None,
#     fragment: str = None,
#     failsafe: bool = False,
# ) -> urllib.parse.ParseResult:
#     """new named tuple from uri with requested part updated"""
#     try:
#         username = first(username, uri.username, "")
#         password = first(password, uri.password, "")
#         hostname = first(hostname, uri.hostname, "")
#         port = first(port, uri.port, "")
#         netloc = (
#             f"{username}{':' if password else ''}{password}"
#             f"{'@' if username or password else ''}{hostname}"
#             f"{':' if port else ''}{port}"
#         )
#         return urllib.parse.urlparse(
#             urllib.parse.urlunparse(
#                 (
#                     first(scheme, uri.scheme),
#                     netloc,
#                     first(path, uri.path),
#                     first(params, uri.params),
#                     first(query, uri.query),
#                     first(fragment, uri.fragment),
#                 )
#             )
#         )
#     except Exception as exc:
#         if failsafe:
#             logger.error(
#                 f"Failed to rebuild "  # lgtm [py/clear-text-logging-sensitive-data]
#                 f"URI {uri} with {scheme=} {username=} {password=} "
#                 f"{hostname=} {port=} {path=} "
#                 f"{params=} {query=} {fragment=} - {exc}"
#             )
#             return uri
#         raise exc


def get_version_ident_for(url: str) -> str:
    """~version~ of the URL data to use for comparisons. Built from headers"""
    try:
        resp = requests.head(url)
        headers = resp.headers
    except Exception:
        logger.warning(f"Unable to HEAD {url}")
        try:
            _, headers = stream_file(
                url=url,
                byte_stream=io.BytesIO(),
                block_size=1,
                only_first_block=True,
            )
        except Exception:
            logger.warning(f"Unable to query image at {url}")
            return

    for header in ("ETag", "Last-Modified", "Content-Length"):
        if headers.get(header):
            return headers.get(header)

    return "-1"


def setup_s3_and_check_credentials(s3_url_with_credentials):
    logger.info("testing S3 Optimization Cache credentials")
    s3_storage = KiwixStorage(s3_url_with_credentials)
    if not s3_storage.check_credentials(
        list_buckets=True, bucket=True, write=True, read=True, failsafe=True
    ):
        logger.error("S3 cache connection error testing permissions.")
        logger.error(f"  Server: {s3_storage.url.netloc}")
        logger.error(f"  Bucket: {s3_storage.bucket_name}")
        logger.error(f"  Key ID: {s3_storage.params.get('keyid')}")
        logger.error(f"  Public IP: {get_public_ip()}")
        raise ValueError("Unable to connect to Optimization Cache. Check its URL.")
    return s3_storage


# #!/usr/bin/env python

# import requests
# import urllib.request

# import backoff

# from os.path import exists, join
# from os import getcwd, mkdir

# from ifixittozim import logger, LANGS


@contextmanager
def setlocale(name):
    with LOCALE_LOCK:
        saved = locale.setlocale(locale.LC_ALL)
        try:
            yield locale.setlocale(locale.LC_ALL, name)
        finally:
            locale.setlocale(locale.LC_ALL, saved)


def backoff_hdlr(details):
    logger.warning(
        "Backing off {wait:0.1f} seconds after {tries} tries "
        "calling function {target} with args {args} and kwargs "
        "{kwargs}".format(**details)
    )


@backoff.on_exception(
    backoff.expo,
    requests.exceptions.RequestException,
    max_time=16,
    on_backoff=backoff_hdlr,
)
def get_api_content(path, **params):
    full_path = get_url(API_PREFIX + path, **params)
    logger.debug(f"Retrieving {full_path}")
    response = requests.get(full_path)
    json_data = response.json() if response and response.status_code == 200 else None
    return json_data


# @backoff.on_exception(backoff.expo,
#                       urllib.error.URLError,
#                       max_time=16,
#                       on_backoff=backoff_hdlr)
# def get_file_content(url, filename):
#     urllib.request.urlretrieve(url, filename)

# def get_cache_path():
#     cwd = getcwd()
#     cachePath = join(cwd, 'cache');
#     while not exists(cachePath):
#         mkdir(cachePath)
#     for asset_kind in ['categories', 'guides', 'images']:
#         subCachePath = join(cwd, 'cache', asset_kind)
#         while not exists(subCachePath):
#             mkdir(subCachePath)
#         if asset_kind not in ['images']:
#             for lang in LANGS:
#                 subCachePath = join(cwd, 'cache', asset_kind, lang);
#                 while not exists(subCachePath):
#                     mkdir(subCachePath)
#     return cachePath

# def get_assets_path():
#     cwd = getcwd()
#     assetPath = join(cwd, 'assets')
#     return assetPath

# def get_dist_path():
#     cwd = getcwd()
#     dist_path = join(cwd, 'dist');
#     while not exists(dist_path):
#         mkdir(dist_path)
#     for asset_kind in ['categories', 'guides', 'images', 'home']:
#         subCachePath = join(cwd, 'dist', asset_kind)
#         while not exists(subCachePath):
#             mkdir(subCachePath)
#         if asset_kind not in ['images']:
#             for lang in LANGS:
#                 subCachePath = join(cwd, 'dist', asset_kind, lang);
#                 while not exists(subCachePath):
#                     mkdir(subCachePath)
#     return dist_path


def guides_in_progress(guides, in_progress=True):
    if in_progress:
        return [guide for guide in guides if "GUIDE_IN_PROGRESS" in guide["flags"]]
    else:
        return [guide for guide in guides if "GUIDE_IN_PROGRESS" not in guide["flags"]]


def get_image_path(image_url):
    return f"../{Global.imager.defer(url=image_url)}"


def _get_image_url_search(obj, for_guide=False):
    if "standard" in obj:
        return obj["standard"]
    elif "medium" in obj:
        return obj["medium"]
    elif "large" in obj:
        return obj["large"]
    elif "original" in obj:
        return obj["original"]
    elif for_guide:
        return DEFAULT_GUIDE_IMAGE_URL
    else:
        raise ImageUrlNotFound(f"Unable to find image URL in object {obj}")


def get_image_url(obj, for_guide=False):
    if "image" in obj and obj["image"]:
        return _get_image_url_search(obj["image"], for_guide)
    else:
        return _get_image_url_search(obj, for_guide)
